__description__ = "a python module housing classifier data specifications"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

from collections import defaultdict
from collections import deque
from future.utils import string_types
import glob
import os
import sys

import h5py
import numpy as np
import pluggy
import toml

from ligo.segments import segment, segmentlist

from ... import configparser
from ... import exceptions
from ... import features
from ... import names
from ... import utils


#-------------------------------------------------
### classifier data specifications

class ClassifierData(object):
    """
    an object which represents the "data" used within classifiers.
    specifically, this object should define getters for each feature that could be requested and store them as @property values.
    In this way, we can encapsulate all the interactions with (S)MDS within this class and maintain a single API with classifiers even if (S)MDS changes (e.g. filesystem query vs remote data base query)

    Note, this is meant as a parent class and specific look-up can be defined through children
    """
    _default_columns = [] ### which columns will be pulled from triggers by default
    _allowed_columns = None ### which columns are allowed; None allows anything
    _required_kwargs = [] ### which kwargs are required for this type of data object

    def __init__(self, start, end, segs=None, columns=None, **kwargs):
        """
        each instance points to (what should be) immutable
            - segments
            - columns
        and when data is queried through .triggers() it will automatically be within these segments and only those columns
        In particular, data is stored internally as a dictionary of the following form
            (channel, numpy.ndarray([np.ndarray([col1_val, col2_val, ...]), ...]))
        the values within the array are ordered according to columns

        *WRITE ME*
            describe attribute inheritence and rational for these
                _default_columns (I want to make each child class declare these for themselves so the developer knows what he's doing)
                _allowed_columns
                _required_kwargs
                _data
                _start
                _end
                _segs
                kwargs
                _columns
            describe method inheritance, what each method is supposed to do
                triggers
                _retrieve_triggers (must be overwritten by child)
                pop
                __add__
        """
        self._data = {}
        self._cached = {}

        assert start <= end, "start must be <= end"
        self._start = start
        self._end = end

        if segs is not None:
            assert utils.livetime(utils.segments_intersection(segmentlist([segment(self._start, self._end)]), segs)) == utils.livetime(segs), '''\
segments must be a subset of [[start, end]] times'''

        else:
            segs = segmentlist([segment(start, end)])
        self._segs = segs

        for kwarg in self._required_kwargs:
            assert kwarg in kwargs, 'kwarg=%s required'%kwarg
        self.kwargs = kwargs

        if not columns:
            columns = self._default_columns

        if self._allowed_columns is not None:
            for col in columns:
                assert col in self._allowed_columns, 'column=%s not allowed'%col
        self._columns = [str(col) for col in columns]

    def is_cached(self, channel, bounds=None):
        """
        Returns whether or not data is cached for the channel requested.
        """
        if bounds is None:
            bounds = dict()

        cached = False
        if channel in self._cached:
            cache = self._cached[channel]
            ### check to see whether there are any cached column bounds that are not within the new requested bounds
            for col in cache.keys():
                if col not in bounds: ### we're requesting possibly looser bounds than we have cached
                   break

                else: ### they share this column, so check relative bounds
                    requested_min, requested_max = bounds[col]
                    cached_min, cached_max = cache[col]
                    if (requested_min < cached_min) or (requested_max > cached_max): ### we want more than is cached
                        break

            else: ### no-one broke, so we're good
                cached = True

        ### either not in the dictionary to begin with or we broke somewhere within the loop over columns
        return cached

    def triggers(self, channels=None, columns=None, segs=None, time=None, bounds=None, **kwargs):
        """
        Submits a query to get features and returns the result of such a query,
        as a dictionary of the following form:

            {channel: numpy.ndarray([np.ndarray([col1_val, col2_val, ...]), ...])}

        The values within the array are ordered according to columns.
        """
        ### check that all columns in bounds will be returned
        if bounds is None:
            bounds = dict()
        else:
            for col in bounds.keys():
                assert col in self.columns, 'bounds can only be placed on columns included in this ClassifierData. column=%s is not'%col

        ### format and validate query
        if channels:
            if isinstance(channels, string_types):
                channels = [channels]

        if segs:
            assert segs in self.segs, 'segments=%s passed in query are not a subset of classifier data segments'%segs
            assert time, 'if passing in segments, time column needs to be specified'

        if columns and (self._allowed_columns is not None):
            if isinstance(columns, string_types):
                columns = [columns]
            for col in columns:
                assert col in self._allowed_columns, 'column=%s not allowed'%col

        ### determine if there's any cached data
        if self._data and (channels is not None):
            channels2query = [chan for chan in channels if not self.is_cached(chan, bounds=bounds)]
            if channels2query:
                self._data.update(self._retrieve_triggers(channels=channels2query, bounds=bounds, **kwargs))

        else: ### no cached data, retrieve all features from query
            self._data.update(self._retrieve_triggers(channels=channels, bounds=bounds, **kwargs))

        ### grab queried data from cache
        if channels:
            queried_data = {channel: self._data[channel] for channel in channels}
        else:
            queried_data = dict(self._data)

        if segs:
            for channel in queried_data.keys():
                times2query = utils.times_in_segments(queried_data[channel][time], segs)
                queried_data[channel] = queried_data[channel][times2query]

        if columns:
            for channel in queried_data.keys():
                queried_data[channel] = queried_data[channel][columns]

        for channel in self._data.keys(): ### update what is considered cached
            if not self.is_cached(channel, bounds=bounds): ### we requested more data than was available
                self._cached[channel] = bounds ### set the bounds to the most recent (widest) request

        for chan, datum in queried_data.items(): ### actually filter this by bounds to be safe (filter cached data)
            queried_data[chan] = utils.filter_triggers(datum, bounds=bounds)

        return queried_data

    @property
    def children(self):
        """
        Makes an iteration over both ClassifierDatas and UmbrellaClassifierDatas a little easier.

        WARNING: Users shouldn't mess with this unless they know what they're doing.
        """
        return [self]

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def segs(self):
        return self._segs

    @property
    def columns(self):
        return self._columns

    @property
    def channels(self):
        return sorted(self._cached.keys())

    def restrict_segments(self, new_segs, time=features.DEFAULT_TIME_NAME):
        """
        update the internal segments to new_segs (checks that new_segs is a strict subset of self.segs)
        iterates through cached triggers and throws away everything outside of new_segs
        NOTE: this requires knowledge of the "time" key within triggers
        """
        assert utils.livetime(utils.segments_intersection(self.segs, new_segs)) == utils.livetime(new_segs), 'new_segs must be a strict subset of existing segs\nsegs=%s\nnew_segs=%s'%(self.segs, new_segs)

        self._segs = segmentlist([segment(s, e) for s, e in new_segs]) ### make a copy so there is no cross-refernce confusion
        for channel in self._data.keys():
            trgs = self._data[channel]
            truth = utils.times_in_segments(trgs[time], self.segs)
            self._data[channel] = trgs[truth]

    def _retrieve_triggers(self, **kwargs):
        raise NotImplementedError('Child classes should define this for themselves!')

    def pop(self, channel, *args):
        """
        Remove (and return) the value associated with channel in the internal structure.

        NOTE: once this is called, object will perform a new query if channel is requested again via triggers.
        """
        self._cached.pop(channel, None)
        return self._data.pop(channel, *args)

    def __add__(self, other):
        """
        return a NEW OBJECT that represents the addition of 2 data objects
        we take the union of segments and columns
        we also keep the union of channels from each object.
        if data was already present in either object, we go ahead and query for any missing data at the time of the addition
        """
        raise NotImplementedError("this could be useful, but could also be difficult to implement. For instance, we need to make sure the children instantiate the correct type of object")

    def target_times(self, time, target_channel, target_bounds, segs=None):
        """
        A convenience function to extract target_times from umbrellas in streaming-based processes.
        """
        ### find target times
        target_times = self.triggers(channels=target_channel, bounds=target_bounds)[target_channel][time]

        ### filter times by segments
        if segs is not None:
            target_times = target_times[utils.times_in_segments(target_times, segs)]

        return target_times

    def random_times(self, time, target_channel, dirty_bounds, dirty_window, random_rate, segs=None):
        """
        A convenience function to extract target_times from umbrellas in streaming-based processes.
        """
        ### draw random times
        dirty_times = self.triggers(channels=target_channel, bounds=dirty_bounds)[target_channel][time]
        dirty_seg = utils.times2segments(dirty_times, dirty_window) ### generate segments

        ### only draw from segments that are outside of dirtysegs but within classifier_data.segs
        random_times = utils.draw_random_times(utils.remove_segments(self.segs, dirty_seg), rate=random_rate)

        ### filter times by segments
        if segs is not None:
            random_times = random_times[utils.times_in_segments(random_times, segs)]

        ### find segments where data is clean
        clean_seg = utils.remove_segments(self.segs, dirty_seg)
        if segs is not None:
            clean_seg = utils.segments_intersection(clean_seg, segs)

        return random_times, clean_seg


class UmbrellaClassifierData(ClassifierData):
    """
    a wrapper around a list of ClassifierData objects, useful when trying to not repeat I/O but splitting it into many small chunks
    """
    def __init__(self, *args, **kwargs):
        ClassifierData.__init__(self, *args, **kwargs)
        self._children = deque(maxlen=kwargs.get('max_children', None))
        self._channels = set()

    @property
    def all_channels(self):
        """
        return all channels present in any child
        """
        chans = set()
        for child in self.children:
            chans |= set(child.channels)
        return list(chans)

    @property
    def columns(self):
        """
        return the columns that are common to all children
        """
        n = len(self.children)
        if n:
            cols = set(self.children[0].columns)
            i = 1
            while i < n:
                cols &= set(self.children[i].columns)
                i += 1
            return list(cols)
        else:
            return []

    def all_columns(self):
        """
        return all columns present in any child
        """
        cols = set()
        for child in self.children:
            cols |= set(child.columns)
        return list(cols)

    @property
    def children(self):
        return self._children

    @property
    def channels(self):
        return list(self._channels)

    def _check_child(self, child):
        """
        make sure the child is not requesting any data not covered by the Umbrella
        """
        lvtm1 = utils.livetime(utils.segments_intersection(self.segs, child.segs))
        lvtm2 = utils.livetime(child.segs)
        assert lvtm1==lvtm2, 'livetime from "self.segs AND child.segs" (%.6f) is not the same as "child.segs" (%.6f)\n%s\n%s'%(lvtm1, lvtm2, utils.segments_intersection(self.segs, child.segs), child.segs)
        ### NOTE:
        ###     Fe could check a lot more there.
        ###     For example, we could check that no children overlap, but we may want children to overlap at some point
        ###         (e.g.: read in both KW triggers and GSTLAL triggers)
        ###     we may also want a check that the full set of children included covers the requested range, but that seems like we should pass children at instantiation time rather than appending htem later. Could be a useful safety net, but not necessarily what we want

    def _update_channels(self, child):
        """
        update Umbrella with common set of channels covered
        """
        if not self._channels:
            self._channels = set(child.channels)
        else:
            self._channels &= set(child.channels)

    def extend_child(self, child):
        """
        modifies self.start, self.end in addition to adding the child
        allways appends child
        """
        self._start = min(self.start, child.start)
        self._end = max(self.end, child.end)
        self._segs = self._segs | child.segs
        self._update_channels(child)
        self._children.append(child)

    def extend_children(self, children):
        """
        modifies self.start, self.end in addition to adding the child
        always appends child
        """
        for child in children:
            self.extend_child(child)

    def strip_child(self):
        """
        removes the left-most child and updates start, end, segs accordingly
        """
        self._children.popleft()
        if len(self.children):
            self._start = min([child.start for child in self._children])
            self._end = max([child.end for child in self._children])
            self._segs = utils.segments_union(*[child.segs for child in self._children])
            self._channels = set()
            for child in self._children:
                self._channels &= set(child.channels)
        else:
            self._start = self._end ### we've removed the last child from the left, so we collapse the segment accordingly
            self._segs = segmentlist([])
            self._channels = set()

    def restrict_segments(self, new_segs, time=features.DEFAULT_TIME_NAME):
        for child in self.children:
            child.restrict_segments(utils.segments_intersection(new_segs, child.segs), time=time)

    def triggers(self, channels=None, columns=None, segs=None, time=None, bounds=None, **kwargs):
        if bounds is None:
            bounds = dict()
        data = {}
        for child in self.children: ### do I/O for each child
            ### trim segs to fit child
            if segs is not None:
                child_segs = utils.segments_intersection(segmentlist([segment(child.start, child.end)]), segs)
            else:
                child_segs = None

            for key, val in child.triggers(channels=channels, columns=columns, segs=child_segs, time=time, bounds=bounds, **kwargs).items():
                if key in data:
                    data[key] = np.concatenate((data[key], val))
                else:
                    data[key] = val

            self._update_channels(child)

        for channel in self.channels: ### update what is considered cached
            if not self.is_cached(channel, bounds=bounds): ### we requested more data than was available
                self._cached[channel] = bounds ### set the bounds to the most recent (widest) request

        return data

    def pop(self, channel, *args):
        """
        Remove (and return) triggers associated with the corresponding channel.
        """
        self._cached.pop(channel, None)
        data = []
        for child in self.children:
            data.append(child.pop(channel, *args))
        return np.concatenate(data)


class MultiSourceClassifierData(UmbrellaClassifierData):
    """a classifier data that gobbles up multiple types of data
    """
    _default_columns = []
    _allowed_columns = None ### allow anything
    _required_kwargs = ['config', 'column_map']

    def __init__(self, *args, **kwargs):
        assert 'columns' in kwargs, '"columns" is required for MultiSourceClassifierData'
        UmbrellaClassifierData.__init__(self, *args, columns=kwargs.pop('columns'), **kwargs)

        ### load classifier data config
        with open(self.kwargs["config"], "r") as f:
            config = toml.load(f)

        factory = ClassifierDataFactory()
        for trg_config in config['data_discovery']: ### iterate over specified ClassifierDatas
            trg_config.update(self.kwargs) ### overwrite as needed
            self.extend_child(factory(self.start, self.end, segs=self.segs, flavor=section, **trg_config))

        ### make sure the columns from each child have an appropriate mapping
        for child in self.children:
            dest = []
            for column in child.columns:
                if column not in self.kwargs['column_map']: ### we require a mapping for each child column
                    raise KeyError('must supply a mapping for each child column into a resulting column. Missing destination for '+column)
                elif self.kwargs['column_map'][column] not in self.columns:
                    raise ValueError('each column must be mapped into one of the delcared destiation column. This is not true for '+column)
                elif self.kwargs['column_map'][column] in dest:
                    raise ValueError('multiple columns cannot be mapped into the same destination. Repeated mapping into '+self.kwargs['column_map'][column])

    @property
    def columns(self):
        return self._columns

    @property
    def column_map(self):
        return self.kwargs['column_map']

    @property
    def invcolumn_map(self):
        ans = defaultdict(list)
        for source, target in self.kwargs['column_map'].items():
            ans[target].append(source)
        return ans

    def triggers(self, channels=None, columns=None, segs=None, time=None, bounds=None, **kwargs):
        if bounds is None:
            bounds = dict()
        data = {}
        column_map = self.column_map
        invcolumn_map = self.invcolumn_map
        if columns is None:
            columns = self.columns

        for child in self.children: ### do I/O for each child
            ### figure out mapping for "time" column
            if time is None:
                child_time = None
            else:
                for child_time in invcolumn_map[time]:
                    if child_time in child.columns:
                        break
                else:
                    raise ValueError('could not find a child column corresponding to '+time)

            ### figure out ordering for the rest of them
            names = []
            for column in columns:
                for child_column in invcolumn_map[column]:
                    if child_column in child.columns:
                        break
                else:
                    raise ValueError('could not find a child column correpsonding to '+column)
                names.append(child_column)

            for key, val in child.triggers(channels=channels, columns=names, segs=segs, time=child_time, bounds=bounds, **kwargs).items(): ### always get all columns
                ### map columns from val into destinations
                val = np.array(zip(*[val[name] for name in names]), dtype=[(column_map[name], val.dtype[name]) for name in names])

                ### store the result
                if key in data:
                    data[key] = np.concatenate((data[key], val))
                else:
                    data[key] = val

        for channel in self.channels: ### update what is considered cached
            if not self.is_cached(channel, bounds=bounds): ### we requested more data than was available
                self._cached[channel] = bounds ### set the bounds to the most recent (widest) request

        return data

    def pop(self, channel, *args):
        """
        Remove (and return) triggers associated with the corresponding channel.
        """
        raise NotImplementedError


#-------------------------------------------------
### plugin specifications

hookspec = pluggy.HookspecMarker("iDQ")

@hookspec
def get_classifier_datum():
    """
    This hook is used to return ClassifierData in the form:
        {"type[:format]": ClassifierData}

    where the type refers to a specific data backend (snax, kleine-welle, etc.)
    and format (optional) refers to a spefific file format and/or data layout.

    """


#-------------------------------------------------

class ClassifierDataFactory(object):
    """
    A factory to create instances of ClassifierData-like objects
    """
    def __init__(self):

        ### set up plugin manager
        manager = pluggy.PluginManager('iDQ')
        manager.add_hookspecs(sys.modules[__name__])

        ### load in reporters by module
        ### FIXME: do this with importlib instead when python3 only
        from . import kw, omicron, snax, synthetic
        manager.register(kw)
        manager.register(omicron)
        manager.register(snax)
        manager.register(synthetic)

        ### add all registered plugins to registry
        self._classifier_datum = {}
        for trigger_plugins in manager.hook.get_classifier_datum():
            self._classifier_datum.update(trigger_plugins)

    def __call__(self, start, end, segs=None, flavor=None, **kwargs):
        """
        standardizes how we map classifier parameters into an actual object

          * start is a float (gps seconds)
          * end is a float (gps seconds)
          * segs, if provided, should be a list of segments (a list of 2-element lists)
          * kwargs should be the result of **config.items(nickname)

        NOTE: 'flavor' is a reserved option that must be supplied

        returns

          * a classifier_data object corresponding to the parameters within **kwargs
        """
        if flavor not in self._classifier_datum:
            raise KeyError('%s is not a known ClassifierData!'%flavor) ### format error reporting a bit better

        return self._classifier_datum[flavor](start, end, segs=segs, **kwargs)
