__description__ = "a python module housing SNAX-based classifier datum"
__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

from collections import defaultdict
import os
import time
import timeit

import h5py
import numpy as np

from ligo.segments import segment, segmentlist

from ... import configparser
from ... import exceptions
from ... import features
from ... import names
from ... import utils
from ... import hookimpl
from . import ClassifierData


#-------------------------------------------------
### classifier data implementations


class SNAXClassifierData(ClassifierData):
    """
    An extension of ClassifierData specifically for extracting SNAX features from hdf5 files.

    We assume the following directory structure:
        ${rootdir}/${basename}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.h5

    """
    _default_columns = utils.GSTLAL_COLUMNS
    _allowed_columns = utils.GSTLAL_COLUMNS
    _required_kwargs = ['rootdir', 'basename']

    def __init__(self, start, end, segs=None, columns=None, **kwargs):
        ClassifierData.__init__(self, start, end, segs=segs, columns=columns, **kwargs)

        ### load cache, or create one if one is not provided
        populate_cache = False
        if 'cache' in self.kwargs and isinstance(self.kwargs['cache'], list):
            if self.kwargs['cache']:
                cache_extent = segment(self.kwargs['cache'][0].segment[0], self.kwargs['cache'][-1].segment[1])
                if utils.livetime(self.segs) > 0 and not (self.segs.extent() in cache_extent):
                    populate_cache = True ### zero livetime or cache doesn't cover span of classifier data
            else:
                populate_cache = True ### cache is empty
        elif 'cache' in self.kwargs:
            self.kwargs['cache'] = utils.file2cache(self.kwargs['rootdir'], self.kwargs['cache'])
        else:
            populate_cache = True

        if populate_cache:
            path = names.basename2gstlal_filedir(self.kwargs['basename'])
            filename = names.basename2gstlal_filename(self.kwargs['basename'])
            self.kwargs['cache'] = utils.path2cache(self.kwargs['rootdir'], os.path.join(path, filename))

        ### filter cache with segments
        self.kwargs['cache'] = [entry for entry in self.kwargs['cache'] if self.segs.intersects_segment(entry.segment)]

    def restrict_segments(self, new_segs, time=features.DEFAULT_TIME_NAME):
        ClassifierData.restrict_segments(self, new_segs, time=time) ### delegate to the parent
        self.kwargs['cache'] = [entry for entry in self.kwargs['cache'] if self.segs.intersects_segment(entry.segment)] ### restrict the cache

    def _retrieve_triggers(self, channels=None, bounds=None, verbose=False, **kwargs):
        """
        workhorse data discovery method for gstlal hdf5 files
        we assume the following directory structure
            ${rootdir}/${basename}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.h5
        """
        if bounds is None:
            bounds = dict()

        if verbose:
            print( '\n'.join([
                '',
                'searching for gstlal triggers within',
                '    rootdir : {rootdir}'.format(**self.kwargs),
                'with',
                '    basename : {basename}'.format(**self.kwargs),
                ])
            )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    {:3f} - {:3f}'.format(*seg) )

        ### data structure and dtype for query results
        data = defaultdict(list)
        dtype = [(col, 'float') for col in self.columns]

        for entry in self.kwargs['cache']: ### iterate over hdf5 files

            ### determine which segments are relevant for this hdf5 file
            trg_seg = segment(names.path2start_end(entry.path, suffix='h5'))
            germane_segments = utils.segments_intersection(self.segs, segmentlist([trg_seg]))

            if verbose:
                print( '        '+entry.path )

            with h5py.File(entry.path, 'r') as file_obj:

                ### determine stride + sample rate of datasets
                assert 'sample_rate' in file_obj.attrs, 'sample_rate metadata not present in hdf5 file'
                sample_rate = file_obj.attrs.get('sample_rate')


                ### query only channels in file
                file_channels = set(file_obj.keys())
                if channels is not None:
                    channels2query = file_channels & set(channels)
                else:
                    channels2query = file_channels - set(self.channels)

                for channel in channels2query:
                    datasets = list(set(file_obj[channel].keys()))
                    dset_segs = [segment(*utils.start_dur2start_end(dset, type_=float)) for dset in datasets]

                    ### filter datasets with segments
                    relevant_datasets = [(dset, seg) for dset, seg in zip(datasets, dset_segs) if germane_segments.intersects_segment(seg)]
                    dset_span = utils.livetime(segmentlist([seg for _, seg in relevant_datasets]))

                    if relevant_datasets:
                        key = os.path.join(channel, relevant_datasets[0][0])
                        row_dtype = file_obj[key].dtype

                        ### preallocate space and initialize all values to nan
                        rows = np.empty((int(dset_span * sample_rate),), dtype=dtype)
                        if rows.size == 1: ### NOTE: workaround for older versions of numpy not able to find ndarray size with a single row
                            rows[:] = tuple(np.nan for col in self.columns)
                        else:
                            rows[:] = np.nan

                        ### populate allocated rows with datasets
                        idx_start = 0
                        for dataset, seg in relevant_datasets:
                            dataset_size = int(utils.livetime(segmentlist([seg])) * sample_rate) - 1
                            key = os.path.join(channel, dataset)

                            rows[idx_start:idx_start+dataset_size] = np.array(file_obj[key])[self.columns]
                            idx_start += dataset_size

                        ### filter after reading everything from this file
                        rows = utils.filter_triggers(rows, bounds=bounds)
                        trimmed_times = utils.times_in_segments(rows[utils.GSTLAL_TIMECOL], germane_segments)
                        data[channel].append(rows[trimmed_times])

        ### stack datasets vertically
        for channel, datasets in data.items():
            data[channel] = np.concatenate(datasets, axis=0)

        ### fill in missing channels
        if channels:
            for channel in channels:
                if channel not in data:
                    data[channel] = np.array([], dtype=dtype)

        return data


class KafkaClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for ingesting SNAX features from Kafka topics.

    Intended to keep a running current timestamp, and has the ability to poll for new data
    and fill its own ClassifierData objects for use when triggers are to be retrieved.

    NOTE: when called, this will cache all trigger regardless of the bounds. This is done to avoid issues with re-querying data from
    rolling buffers, which is not guaranteed to return consistent results. Instead, we record everything we query and filter within 
    ClassifierData.triggers()
    """
    _default_columns = utils.GSTLAL_COLUMNS
    _allowed_columns = utils.GSTLAL_COLUMNS
    _required_kwargs = ['group', 'port', 'topic', 'poll_timeout', 'retry_cadence', 'sample_rate']

    def __init__(self, *args, **kwargs):
        ClassifierData.__init__(self, *args, **kwargs)

        ### link to Kafka consumer if one has not already been assigned
        if 'CONSUMER' not in self.kwargs:
            ### check if Kafka is installed
            utils.check_kafka()

            ### create a Kafka consumer
            self.kwargs['CONSUMER'] = utils.KafkaConsumer(
                self.kwargs['group'],
                self.kwargs['port'],
                topics=[self.kwargs['topic']],
                **self.kwargs.get('kafka_kwargs', {})
            )

            ### assign to specific partition to allow seeks/consumes
            self.kwargs['CONSUMER'].assign([
                utils.kafka.TopicPartition(self.kwargs['topic'], partition=self.kwargs['CONSUMER'].partition),
            ])

        ### dtype for query results
        self._dtype = [(col, 'float') for col in self.columns]

    def _retrieve_triggers(self, channels=None, verbose=False, **kwargs):
        """NOTE: we intentionally ignore bounds here and just store everything
        """
        if not self._data: ### NOTE: this forces data to only be populated once

            if not utils.livetime(self.segs): ### we'll never find any data in the segments because they have zero livetime
                self._data = {}
                if channels is not None:
                    for channel in channels:
                        self._data[channel] = np.array([], dtype=self._dtype)
                return self._data

            ### there is the possibility of finding some data, so we go looking for that
            ### NOTE: assume last processed timestamp is pointed just before this span
            stride = 1. / self.kwargs['sample_rate']
            timestamp = self.start - stride
            segs = segmentlist([segment(self.start, self.end)]) ### the span of this object

            ### attempt to seek to correct offset
            try:
                self.kwargs['CONSUMER'].seek_timestamp(self.kwargs['topic'], self.start, stride)
            except utils.kafka.KafkaException as e:
                print(repr(e))

            ### check if timestamps fall within span of classifier data
            stop = self.end - stride
            while timestamp < stop:

                ### poll for data at requested offset
                start_time = timeit.default_timer()
                new_timestamp, data_buffer = self.kwargs['CONSUMER'].retrieve(timeout=self.kwargs['poll_timeout'])

                if data_buffer:
                    if verbose:
                        print('    timestamp of buffer: %.6f'%new_timestamp)
                        print('    latency: %.6f'%utils.gps2latency(new_timestamp))

                    if new_timestamp == timestamp+stride: ### data is contiguous, which is GOOD
                        if utils.time_in_segments(new_timestamp, self.segs): ### we care about this data

                            ### fill in all data from buffer
                            data = {}
                            for channel, rows in data_buffer['features'].items():
                                data[channel] = np.array([tuple(row[col] for col in self.columns) for row in rows if row], dtype=self._dtype)

                            ### add to existing triggers
                            for channel, triggers in data.items():
                                if channel in self._data:
                                    self._data[channel] = np.concatenate((self._data[channel], triggers))
                                else:
                                    self._data[channel] = triggers

                        else:
                            pass ### we don't care about this data, so we just mosey along

                    else: ### data is NOT contiguous, which could be BAD
                        gap = segment(timestamp+stride, new_timestamp)

                        ### check whether the gap touched things we care about, which is BAD
                        if utils.livetime(utils.segments_intersection(self.segs, segmentlist([gap]))):
                            newer_timestamp = max(utils.floor_div(timestamp, self.end - self.start), self.end) ### move to the next stride
                            self.kwargs['CONSUMER'].seek_timestamp(self.kwargs['topic'], newer_timestamp, stride) ### reset the consumer's position
                            raise exceptions.IncontiguousDataError(gap[0], gap[1], newer_timestamp) ### report the error

                        else:
                            pass ### there is a gap, but it doesn't touch anything we actually want so who cares?

                    timestamp = new_timestamp ### update the timestamp to be the new thing

                ### wait to poll for new data
                elapsed = timeit.default_timer() - start_time
                sleep_time = self.kwargs['retry_cadence'] - min(elapsed, self.kwargs['retry_cadence'])
                time.sleep(sleep_time)

            ### timestamp >= stop
            if self._data: ### if we received the full data requested, return that

                ### filter by segs if needed
                if not utils.livetime(self.segs) == utils.livetime(utils.segments_intersection(self.segs, segs)):
                    for channel, triggers in self._data.items():
                        trimmed_times = utils.times_in_segments(triggers[utils.GSTLAL_TIMECOL], self.segs)
                        self._data[channel] = triggers[trimmed_times]

            else: ### did not find ANY data within the requested time
                raise exceptions.NoDataError(self.start, self.end-self.start)

        ### fill in missing channels
        if channels:
            for channel in channels:
                if channel not in self._data: ### ignore bounds here
                    self._data[channel] = np.array([], dtype=self._dtype)

        ### make sure we record that we cached everything for all channels present
        bounds = {} ### we do not impose any bounds
        for channel in self._data: ### update what is considered cached
            if channel in self._cached: ### not yet cached recoreded as cached
                self._cached[channel] = bounds ### we record that we cached everything for all channels

        return self._data


#-------------------------------------------------
### plugin implementations

@hookimpl
def get_classifier_datum():
    return {
        "snax": SNAXClassifierData,
        "snax:hdf5": SNAXClassifierData,
        "snax:kafka": KafkaClassifierData,
    }
