__description__ = "a module that tests several aspects of the io submodule"
__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#--------------------------------------------------------------------------

import sys
import pytest

import numpy as np

from idq import configparser
from idq import factories
from idq import utils
from idq.io.triggers import snax

#--------------------------------------------------------------------------

#----------------------
#        tests
#----------------------

@pytest.mark.usefixtures('snax_classifierdata')
class TestFactories(object):
    """
    Tests several features within Factory classes.
    """
    def test_classifierdata_factory(self, snax_classifierdata_conf):
        conf = snax_classifierdata_conf
        conf['flavor'] = 'snax:hdf5'

        factory = factories.ClassifierDataFactory()
        start_time = snax_classifierdata_conf['start_time']
        end_time = snax_classifierdata_conf['end_time']
        newclassifierdata = factory(start_time, end_time, **conf)

        ### check that this factory generated object derives from the right class
        ### and has the same properties
        assert isinstance(newclassifierdata, snax.SNAXClassifierData), 'factory-generated ClassifierData not the right class'
        assert newclassifierdata.start == self.classifierdata.start, 'factory-generated ClassifierData start time not set properly'
        assert newclassifierdata.end == self.classifierdata.end, 'factory-generated ClassifierData end time not set properly'
        assert newclassifierdata.segs == self.classifierdata.segs, 'factory-generated ClassifierData segs not set properly'
        assert newclassifierdata.columns == self.classifierdata.columns, 'factory-generated ClassifierData columns not set properly'
        assert newclassifierdata.channels == self.classifierdata.channels, 'factory-generated ClassifierData channels not set properly'

    def test_quiver_factory_labeled_sparse(self, snax_classifierdata_conf):
        target_channel = snax_classifierdata_conf['channels'][0]
        target_bounds = configparser.config2bounds({'snr': [7, 'inf']})
        dirty_bounds = configparser.config2bounds({'snr': [4, 'inf']})
        dirty_window = 0.2
        time = 'time'
        rate = 2

        target_times = utils.filter_triggers(self.classifierdata.triggers(channels=target_channel)[target_channel], bounds=target_bounds)[time]
        dirty_times = utils.filter_triggers(self.classifierdata.triggers(channels=target_channel)[target_channel], bounds=dirty_bounds)[time]
        dirty_seg = utils.times2segments(dirty_times, dirty_window) ### generate segments
        random_times = utils.draw_random_times(utils.remove_segments(self.classifierdata.segs, dirty_seg), rate=rate)

        quiver = factories.QuiverFactory(self.classifierdata).labeled(target_times, random_times)

        assert len(quiver) == (len(target_times) + len(random_times)), 'quiver not the right size'
        assert quiver.segs == self.classifierdata.segs, 'quiver segs not set properly'
        assert quiver.start == self.classifierdata.start, 'quiver start time not set properly'
        assert quiver.end == self.classifierdata.end, 'quiver end time not set properly'

        num_glitch, num_clean = quiver.vectors2num()
        assert num_glitch == len(target_times), 'quiver does not have the right number of glitch samples'
        assert num_clean == len(random_times), 'quiver does not have the right number of clean samples'

    def test_quiver_factory_labeled_dense(self, snax_classifierdata_conf):
        target_channel = snax_classifierdata_conf['channels'][0]
        target_bounds = configparser.config2bounds({'snr': [7, 'inf']})
        dirty_bounds = configparser.config2bounds({'snr': [4, 'inf']})
        dirty_window = 0.2
        time = 'time'
        rate = 0.0625

        target_times = utils.filter_triggers(self.classifierdata.triggers(channels=target_channel)[target_channel], bounds=target_bounds)[time]
        dirty_times = utils.filter_triggers(self.classifierdata.triggers(channels=target_channel)[target_channel], bounds=dirty_bounds)[time]
        dirty_seg = utils.times2segments(dirty_times, dirty_window) ### generate segments
        random_times = utils.draw_random_times(utils.remove_segments(self.classifierdata.segs, dirty_seg), rate=rate)

        quiver = factories.QuiverFactory(self.classifierdata).labeled(target_times, random_times)

        assert len(quiver) == (len(target_times) + len(random_times)), 'quiver not the right size'
        assert quiver.segs == self.classifierdata.segs, 'quiver segs not set properly'
        assert quiver.start == self.classifierdata.start, 'quiver start time not set properly'
        assert quiver.end == self.classifierdata.end, 'quiver end time not set properly'

        num_glitch, num_clean = quiver.vectors2num()
        assert num_glitch == len(target_times), 'quiver does not have the right number of glitch samples'
        assert num_clean == len(random_times), 'quiver does not have the right number of clean samples'

    def test_quiver_factory_unlabeled_sparse(self, snax_classifierdata_conf):
        dt = 2
        quiver = factories.QuiverFactory(self.classifierdata).unlabeled(dt=dt)

        assert len(quiver) == (utils.livetime(self.classifierdata.segs) / 2.)
        assert quiver.segs == self.classifierdata.segs, 'quiver segs not set properly'
        assert quiver.start == self.classifierdata.start, 'quiver start time not set properly'
        assert quiver.end == self.classifierdata.end, 'quiver end time not set properly'

        for vector in quiver:
            assert not vector.label, 'feature vector should be unlabeled'
            assert not vector.rank, 'feature vector should be unranked'

    def test_quiver_factory_unlabeled_dense(self, snax_classifierdata_conf):
        dt = 0.0625
        quiver = factories.QuiverFactory(self.classifierdata).unlabeled(dt=dt)

        assert len(quiver) == (utils.livetime(self.classifierdata.segs) / 0.0625)
        assert quiver.segs == self.classifierdata.segs, 'quiver segs not set properly'
        assert quiver.start == self.classifierdata.start, 'quiver start time not set properly'
        assert quiver.end == self.classifierdata.end, 'quiver end time not set properly'

        for vector in quiver:
            assert not vector.label, 'feature vector should be unlabeled'
            assert not vector.rank, 'feature vector should be unranked'

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_quiver_factory_labeled_windowed(self, snax_classifierdata_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_quiver_factory_unlabeled_windowed(self, snax_classifierdata_conf):
    #    raise NotImplementedError
