__description__ = "a module that tests several aspects of the Quiver class"
__author__ = "Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#--------------------------------------------------------------------------

import sys
import pytest

import numpy as np

from idq import configparser
from idq import factories
from idq import utils

#--------------------------------------------------------------------------

#----------------------
#        tests
#----------------------

@pytest.mark.usefixtures('snax_quiver')
class TestLabeledQuiver(object):
    """
    Tests several features for a labeled Quiver.
    """
    def test_times(self, snax_quiver_conf):
        assert len(self.labeled_quiver) == (len(self.target_times) + len(self.random_times)), 'quiver not the right size'
        assert self.labeled_quiver.segs == self.classifierdata.segs, 'quiver segs not set properly'
        assert self.labeled_quiver.start == self.classifierdata.start, 'quiver start time not set properly'
        assert self.labeled_quiver.end == self.classifierdata.end, 'quiver end time not set properly'

    def test_coverage(self, snax_quiver_conf):
        self.labeled_quiver.check_coverage()

    def test_vectors2classes(self, snax_quiver_conf):
        glitch, clean = self.labeled_quiver.vectors2classes()

        assert len(glitch) == len(self.target_times), 'wrong number of glitch samples in quiver'
        assert len(clean) == len(self.random_times), 'wrong number of clean samples in quiver'

        for vector in glitch:
            assert vector.label == 1, 'wrong label for glitch sample'
        for vector in clean:
            assert vector.label == 0, 'wrong label for clean sample'

    def test_vectors2num(self, snax_quiver_conf):
        num_glitch, num_clean = self.labeled_quiver.vectors2num()
        assert num_glitch == len(self.target_times), 'quiver does not have the right number of glitch samples'
        assert num_clean == len(self.random_times), 'quiver does not have the right number of clean samples'

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_remove_vectors(self, snax_quiver_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_copy(self, snax_quiver_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="need numpy 1.9.0+ to use np.unique with return_counts kwarg")
    #def test_labels(self, snax_quiver_conf):
    #    unique, counts = np.unique(self.labeled_quiver.labels, return_counts=True)
    #    counter = dict(zip(unique, counts))

    #    assert set(counter.keys()) == set([0, 1]), 'more labels in quiver than allowed'
    #    assert len(counter[0]) == len(self.random_times), 'quiver.labels does not have the right number of clean samples'
    #    assert len(counter[1]) == len(self.target_times), 'quiver.labels does not have the right number of clean samples'

    def test_ranks(self, snax_quiver_conf):
        is_none = self.labeled_quiver.ranks == np.array(None)
        assert np.all(is_none), 'quiver.ranks contain a quantity other than None'

    def test_gpss(self, snax_quiver_conf):
        times = np.concatenate((self.target_times, self.random_times)).sort()
        assert np.array_equal(self.labeled_quiver.gpss.sort(), times), 'gps times of quiver do not match times used to generate quiver'

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_apply_ranks(self, snax_quiver_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_append_quiver(self, snax_quiver_conf):
    #    raise NotImplementedError

    def test_serial_vectorize(self, snax_quiver_conf):
        result = self.labeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, time=snax_quiver_conf['time'])

    def test_parallel_vectorize(self, snax_quiver_conf):
         num_procs = [2, 4, 8]
         for num_proc in num_procs:
             result = self.labeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, num_proc=num_proc)

    def test_check_serial_parallel_vectorize_match(self, snax_quiver_conf):
        serial_data, serial_names = self.labeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, time=snax_quiver_conf['time'])

        num_procs = [2, 4, 8]
        for num_proc in num_procs:
            parallel_data, parallel_names = self.labeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, num_proc=num_proc)
            assert np.array_equal(serial_data.sort(), parallel_data.sort()), 'serial and parallel vectorize() data do not match'

@pytest.mark.usefixtures('snax_quiver')
class TestUnlabeledQuiver(object):
    """
    Tests several features for an unlabeled Quiver.
    """
    def test_properties(self, snax_quiver_conf):
        assert len(self.unlabeled_quiver) == (utils.livetime(self.classifierdata.segs) / 2.)
        assert self.unlabeled_quiver.segs == self.classifierdata.segs, 'quiver segs not set properly'
        assert self.unlabeled_quiver.start == self.classifierdata.start, 'quiver start time not set properly'
        assert self.unlabeled_quiver.end == self.classifierdata.end, 'quiver end time not set properly'

        for vector in self.unlabeled_quiver:
            assert not vector.label, 'feature vector should be unlabeled'
            assert not vector.rank, 'feature vector should be unranked'

    def test_coverage(self, snax_quiver_conf):
        self.unlabeled_quiver.check_coverage()

    def test_vectors2classes(self, snax_quiver_conf):
        glitch, clean = self.unlabeled_quiver.vectors2classes()

        assert len(glitch) == 0, 'glitch samples present in unlabeled quiver'
        assert len(clean) == 0, 'samples present in unlabeled quiver'

    def test_vectors2num(self, snax_quiver_conf):
        num_glitch, num_clean = self.unlabeled_quiver.vectors2num()

        assert num_glitch == 0, 'glitch samples present in unlabeled quiver'
        assert num_clean == 0, 'samples present in unlabeled quiver'

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_remove_vectors(self, snax_quiver_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_copy(self, snax_quiver_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_labels(self, snax_quiver_conf):
    #    raise NotImplementedError

    def test_ranks(self, snax_quiver_conf):
        is_none = self.unlabeled_quiver.ranks == np.array(None)
        assert np.all(is_none), 'quiver.ranks contain a quantity other than None'

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_gpss(self, snax_quiver_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_apply_ranks(self, snax_quiver_conf):
    #    raise NotImplementedError

    #@pytest.mark.skip(reason="not implemented yet")
    #def test_append_quiver(self, snax_quiver_conf):
    #    raise NotImplementedError

    def test_serial_vectorize(self, snax_quiver_conf):
        result = self.unlabeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, time=snax_quiver_conf['time'])

    def test_dense_vectorize(self, snax_quiver_conf):
        result = self.dense_unlabeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, time=snax_quiver_conf['time'])

    def test_parallel_vectorize(self, snax_quiver_conf):
         num_procs = [2, 4, 8]
         for num_proc in num_procs:
             result = self.unlabeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, num_proc=num_proc)

    def test_check_serial_parallel_vectorize_match(self, snax_quiver_conf):
        serial_data, serial_names = self.unlabeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, time=snax_quiver_conf['time'])

        num_procs = [2, 4, 8]
        for num_proc in num_procs:
            parallel_data, parallel_names = self.unlabeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, num_proc=num_proc)
            assert np.array_equal(serial_data.sort(), parallel_data.sort()), 'serial and parallel vectorize() data do not match'

    def test_check_dense_sparse_vectorize_match(self, snax_quiver_conf):
        sparse_data, sparse_names = self.unlabeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, time=snax_quiver_conf['time'])
        dense_data, dense_names = self.dense_unlabeled_quiver.vectorize(snax_quiver_conf['safe_channels'], self.select, time=snax_quiver_conf['time'])

        assert np.array_equal(sparse_data, dense_data), 'regular and dense quiver.vectorize() data do not match'
